// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Ray3D.hpp"

using namespace Dawn;

struct Ray3D {
  glm::vec3 origin;
  glm::vec3 direction;
};

struct PhysicsSphere {
  glm::vec3 center;
  float_t radius;
};

bool_t Dawn::raytestSphere(
  struct Ray3D ray,
  struct PhysicsSphere sphere,
  glm::vec3 *hit,
  glm::vec3 *normal
) {
  glm::vec3 h, n;
  auto result = glm::intersectRaySphere(
    ray.origin, ray.direction,
    sphere.center, sphere.radius,
    h, n
  );

  if(hit != nullptr) *hit = h;
  if(normal != nullptr) *normal = n;

  return result;
}

bool_t Dawn::raytestTriangle(
  struct Ray3D ray,
  struct PhysicsTriangle triangle,
  glm::vec2 *hit,
  float_t *distance
) {
  glm::vec2 h;
  float_t d;
  auto result = glm::intersectRayTriangle(
    ray.origin, ray.direction,
    triangle.v0, triangle.v1, triangle.v2,
    h, d
  );
  if(hit != nullptr) *hit = h;
  if(distance != nullptr) *distance = d;
  return result;
}

bool_t Dawn::raytestAABB(
  struct Ray3D ray,
  struct AABB3D box,
  glm::vec3 *point,
  glm::vec3 *normal,
  float_t *distance
) {
  // Compute the inverse direction of the ray, for numerical stability
  glm::vec3 invDir(1.0f / ray.direction.x, 1.0f / ray.direction.y, 1.0f / ray.direction.z);

  // Compute the t-values for the two intersection candidates
  glm::vec3 tMin = (box.min - ray.origin) * invDir;
  glm::vec3 tMax = (box.max - ray.origin) * invDir;

  // Make sure tMin is less than or equal to tMax for all components
  glm::vec3 t1 = glm::min(tMin, tMax);
  glm::vec3 t2 = glm::max(tMin, tMax);
  float tNear = glm::compMax(t1);
  float tFar = glm::compMin(t2);

  // If tNear is greater than or equal to tFar, there is no intersection
  if (tNear >= tFar) return false;

  // If tFar is negative, the ray is pointing away from the box
  if(tFar < 0.0f) return false;

  // Compute the hit point and normal
  glm::vec3 hitPoint = ray.origin + tNear * ray.direction;

  if(point != nullptr) *point = hitPoint;
  if(distance != nullptr) *distance = tNear;
  if(normal != nullptr) {
    if (hitPoint.x == box.min.x) {
      *normal = glm::vec3(-1, 0, 0);
    } else if (hitPoint.x == box.max.x) {
      *normal = glm::vec3(1, 0, 0);
    } else if (hitPoint.y == box.min.y) {
      *normal = glm::vec3(0, -1, 0);
    } else if (hitPoint.y == box.max.y) {
      *normal = glm::vec3(0, 1, 0);
    } else if (hitPoint.z == box.min.z) {
      *normal = glm::vec3(0, 0, -1);
    } else if (hitPoint.z == box.max.z) {
      *normal = glm::vec3(0, 0, 1);
    }
  }

  // The ray intersects the box
  return true;
}